/* Autogenerated by png2wx.pl on Thu May  1 11:32:28 2014 */
#ifndef ICONS_H
#define ICONS_H 1

#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#	include <wx/wx.h>
#endif

#include <wx/filename.h>
#include <wx/mstream.h>

extern void initialize_images(void);

extern wxBitmap *_img_weatherfax;

#ifdef OCPN_USE_SVG
extern wxString _svg_weatherfax;
extern wxString _svg_weatherfax_rollover;
extern wxString _svg_weatherfax_toggled;
#endif

#endif /* ICONS_H */
